(****** Encoding of A Concurrent Language for Certified Programming ******)
(******              created February 28 2004, revised June 29 2004 ******)

Require EqNat.
Require Max.
Require Lt.
Require Plus.
Require Arith.
Require Export Coq.Lists.PolyList.
Require Export heap2.

(****** Special tactic to use the fsl hints database ******)
Tactic Definition FAuto := Auto with fsl.
Tactic Definition SplitAllHyps :=  (* splits all conjunctions in hypotheses *)
  Repeat Match Context With [H:?/\? |- ?] -> Inversion_clear H.
Tactic Definition SplitAllExs :=   (* Splits all existentials in hypothesis *)
  Repeat Match Context With [H:(ex ? ?) |- ?] -> Inversion_clear H.


(****** Syntactic Categories & Related Utilities ******)

Syntactic Definition lab := nat.		(* code label *)

Inductive reg : Set :=				(* registers *)
   r0 : reg | r1 : reg | r2 : reg | r3 : reg
 | r4 : reg | r5 : reg | r6 : reg | r7 : reg.

Definition beq_reg :=				(* comp r1 r2, ret bool *)
 [r1,r2:reg]
 Cases r1 r2 of
  r0 r0 => true | r1 r1 => true | r2 r2 => true | r3 r3 => true
 | r4 r4 => true | r5 r5 => true | r6 r6 => true | r7 r7 => true
 | _ _ => false
 end.

Inductive rfile : Set :=			(* register file *)
   emptyR : rfile
 | consR : rfile -> reg -> int -> rfile.

Fixpoint lookupR[rf:rfile] : reg -> int :=	(* look up reg file *)
 [r:reg]
 Cases rf of
  emptyR => O
 | (consR rf' r' i) => (if (beq_reg r r') then i
                        else (lookupR rf' r))
 end.

Definition updateR : rfile -> reg -> int -> rfile := (* update reg file *)
 [rf:rfile; r:reg; i:int]
 (consR rf r i).

Syntactic Definition state := (heap * rfile).	(* state *)

Definition getValinH [s:state] : addr -> hvopt := (* get val from H of state *)
 [i:addr]
  let (H, R) = s in (hread H i).

Definition getValinR [s:state] : reg -> int :=	(* get val from R of state *)
 [r:reg]
  let (H, R) = s in (lookupR R r).

Inductive comm : Set :=				(* commands *)
 | add : reg -> reg -> reg -> comm
 | sub : reg -> reg -> reg -> comm
 | movi : reg -> int -> comm.

Inductive iseq : Set :=				(* instruction sequence *)
  seq : comm -> iseq -> iseq
 | bgt : reg -> reg -> lab -> iseq -> iseq
 | be : reg -> reg -> lab -> iseq -> iseq
 | ld : reg -> reg -> nat -> iseq -> iseq
 | st : reg -> nat -> reg -> iseq -> iseq
 | yield : iseq -> iseq
 | jd : lab -> iseq.

Inductive codeheap : Set :=			(* code heap *)
 | emptyC : codeheap
 | consC : lab -> iseq -> codeheap -> codeheap.

Inductive lookupC : codeheap -> lab -> iseq -> Prop :=
  						(* look up code heap *)
 | lookupC_eq : (ch:codeheap; l:lab; b:iseq)
                 (lookupC (consC l b ch) l b)
 | lookupC_neq : (ch:codeheap; l,l':lab; b,b':iseq)
		  (lookupC ch l' b')
		   -> (~l=l')
		   -> (lookupC (consC l b ch) l' b').

Definition prog := ((state * (codeheap * codeheap)) * ((iseq * iseq) * nat)).
						(* program *)

Definition assert := state -> Prop.		(* invariants & asserts *)

Definition guar := state -> state -> Prop.	(* assumptions & guarantees *)

Inductive codeheapty : Type :=			(* code heap type *)
 | emptyCT : codeheapty
 | consCT : lab -> assert -> guar -> codeheapty -> codeheapty.

Inductive lookupCT : codeheapty -> lab -> assert -> guar -> Prop :=
  (* look up code heap type *)
 | lookupCT_eq : (ct:codeheapty; l:lab; p:assert; g:guar)
                 (lookupCT (consCT l p g ct) l p g)
 | lookupCT_neq : (ct:codeheapty; l,l':lab; p,p':assert; g,g':guar)
		  (lookupCT ct l' p' g')
		   -> (~l=l')
		   -> (lookupCT (consCT l p g ct) l' p' g').

Tactic Definition LookupCT := Repeat
 (Apply lookupCT_eq) Orelse (Apply lookupCT_neq; [Auto | Discriminate]).

Definition spec := (prodT
   (prodT assert (prodT codeheapty codeheapty))
   (prodT (prodT guar guar) (prodT guar guar))).
						(* spec *)

Definition getCT1 := [phi:spec] (fstT (sndT (fstT phi))).
Definition getCT2 := [phi:spec] (sndT (sndT (fstT phi))).
Definition getInv := [phi:spec] (fstT (fstT phi)).
Definition getAG1 := [phi:spec] (fstT (sndT phi)).
Definition getAG2 := [phi:spec] (sndT (sndT phi)).

Definition tspec := (prodT
   (prodT assert codeheapty)
   (prodT guar guar)).				(* thread spec *)

Definition tgetI := [t:tspec] (fstT (fstT t)).
Definition tgetCT := [t:tspec] (sndT (fstT t)).
Definition tgetAG := [t:tspec] (sndT t).
Definition tgetA := [t:tspec] (fstT (sndT t)).
Definition tgetG := [t:tspec] (sndT (sndT t)).

Definition getT1 : spec -> tspec
  := [phi:spec] (pairT (pairT (getInv phi) (getCT1 phi)) (getAG1 phi)).
Definition getT2 : spec -> tspec
  := [phi:spec] (pairT (pairT (getInv phi) (getCT2 phi)) (getAG2 phi)).
 
(****** Operational Semantics ******)

Definition Next[c:comm] : state -> state :=	(* auxiliary state update *)
 [s:state] let (H, R) = s in
 Cases c of
   | (add rd rs rt) => (H,(updateR R rd (plus (lookupR R rs) (lookupR R rt))))
   | (sub rd rs rt) => (H,(updateR R rd (minus (lookupR R rs) (lookupR R rt))))
   | (movi rd i) => (H,(updateR R rd i))
 end.

Inductive Step : prog -> prog -> Prop :=
 | S1_seq : (s:state; ch1,ch2:codeheap; c:comm; b1',b2:iseq)
           (Step ((s, (ch1, ch2)), (((seq c b1'), b2), (1)))
	         (((Next c s), (ch1, ch2)), ((b1', b2), (1))))
 | S2_seq : (s:state; ch1,ch2:codeheap; c:comm; b1,b2':iseq)
           (Step ((s, (ch1, ch2)), ((b1, (seq c b2')), (2)))
	         (((Next c s), (ch1, ch2)), ((b1, b2'), (2))))
 | S1_ld : (H:heap; R:rfile; ch1,ch2:codeheap; rd,rs:reg; w,n:nat; b1',b2:iseq)
	   ((hread H (plus w (lookupR R rs))) = (hvsome n))
	    -> (Step (((H, R), (ch1, ch2)), (((ld rd rs w b1'), b2), (1)))
                     (((H, (updateR R rd n)), (ch1, ch2)), ((b1', b2), (1))))
 | S2_ld : (H:heap; R:rfile; ch1,ch2:codeheap; rd,rs:reg; w,n:nat; b1,b2':iseq)
	   ((hread H (plus w (lookupR R rs))) = (hvsome n))
	    -> (Step (((H, R), (ch1, ch2)), ((b1, (ld rd rs w b2')), (2)))
                     (((H, (updateR R rd n)), (ch1, ch2)), ((b1, b2'), (2))))
 | S1_st : (H:heap; R:rfile; ch1,ch2:codeheap; rd,rs:reg; w,n:nat; b1',b2:iseq)
	   (hindom (plus w (lookupR R rd)) H)
	    -> (Step (((H, R), (ch1, ch2)), (((st rd w rs b1'), b2), (1)))
                     ((((hwrite H (plus w (lookupR R rd)) (lookupR R rs)), R),
			 (ch1, ch2)), ((b1', b2), (1))))
 | S2_st : (H:heap; R:rfile; ch1,ch2:codeheap; rd,rs:reg; w,n:nat; b1,b2':iseq)
	   (hindom (plus w (lookupR R rd)) H)
	    -> (Step (((H, R), (ch1, ch2)), ((b1, (st rd w rs b2')), (2)))
                     ((((hwrite H (plus w (lookupR R rd)) (lookupR R rs)), R),
			 (ch1, ch2)), ((b1, b2'), (2))))
 | S1_bgtF : (H:heap; R:rfile; ch1,ch2:codeheap; rs,rt:reg; f:lab; b1',b2:iseq)
	     (le (lookupR R rs) (lookupR R rt))
	      -> (Step (((H, R), (ch1, ch2)), (((bgt rs rt f b1'), b2), (1)))
	               (((H, R), (ch1, ch2)), ((b1', b2), (1))))
 | S2_bgtF : (H:heap; R:rfile; ch1,ch2:codeheap; rs,rt:reg; f:lab; b1,b2':iseq)
	     (le (lookupR R rs) (lookupR R rt))
	      -> (Step (((H, R), (ch1, ch2)), ((b1, (bgt rs rt f b2')), (2)))
	               (((H, R), (ch1, ch2)), ((b1, b2'), (2))))
 | S1_bgtT : (H:heap; R:rfile; ch1,ch2:codeheap; rs,rt:reg; f:lab)
	     (b1',b1'',b2:iseq)
	     (gt (lookupR R rs) (lookupR R rt))
	      -> (lookupC ch1 f b1'')
	      -> (Step (((H, R), (ch1, ch2)), (((bgt rs rt f b1'), b2), (1)))
	               (((H, R), (ch1, ch2)), ((b1'', b2), (1))))
 | S2_bgtT : (H:heap; R:rfile; ch1,ch2:codeheap; rs,rt:reg; f:lab)
	     (b1,b2',b2'':iseq)
	     (gt (lookupR R rs) (lookupR R rt))
	      -> (lookupC ch2 f b2'')
	      -> (Step (((H, R), (ch1, ch2)), ((b1, (bgt rs rt f b2')), (2)))
	               (((H, R), (ch1, ch2)), ((b1, b2''), (2))))
 | S1_beF : (H:heap; R:rfile; ch1,ch2:codeheap; rs,rt:reg; f:lab; b1',b2:iseq)
	    ~((lookupR R rs) = (lookupR R rt))
	     -> (Step (((H, R), (ch1, ch2)), (((be rs rt f b1'), b2), (1)))
	              (((H, R), (ch1, ch2)), ((b1', b2), (1))))
 | S2_beF : (H:heap; R:rfile; ch1,ch2:codeheap; rs,rt:reg; f:lab; b1,b2':iseq)
	    ~((lookupR R rs) = (lookupR R rt))
	     -> (Step (((H, R), (ch1, ch2)), ((b1, (be rs rt f b2')), (2)))
	              (((H, R), (ch1, ch2)), ((b1, b2'), (2))))
 | S1_beT : (H:heap; R:rfile; ch1,ch2:codeheap; rs,rt:reg; f:lab)
	    (b1',b1'',b2:iseq)
	    ((lookupR R rs) = (lookupR R rt))
	     -> (lookupC ch1 f b1'')
	     -> (Step (((H, R), (ch1, ch2)), (((be rs rt f b1'), b2), (1)))
	              (((H, R), (ch1, ch2)), ((b1'', b2), (1))))
 | S2_beT : (H:heap; R:rfile; ch1,ch2:codeheap; rs,rt:reg; f:lab)
	    (b1,b2',b2'':iseq)
	    ((lookupR R rs) = (lookupR R rt))
	     -> (lookupC ch2 f b2'')
	     -> (Step (((H, R), (ch1, ch2)), ((b1, (be rs rt f b2')), (2)))
	              (((H, R), (ch1, ch2)), ((b1, b2''), (2))))
 | S1_yld : (i:nat; s:state; ch1,ch2:codeheap; b1',b2:iseq)
            (i=(1)\/i=(2))
            -> (Step ((s, (ch1, ch2)), (((yield b1'), b2), (1)))
	             ((s, (ch1, ch2)), ((b1', b2), i)))
 | S2_yld : (i:nat; s:state; ch1,ch2:codeheap; b1,b2':iseq)
            (i=(1)\/i=(2))
            -> (Step ((s, (ch1, ch2)), ((b1, (yield b2')), (2)))
	             ((s, (ch1, ch2)), ((b1, b2'), i)))
 | S1_jd : (s:state; ch1,ch2:codeheap; f:lab; b1',b2:iseq)
           (lookupC ch1 f b1')
           -> (Step ((s, (ch1, ch2)), (((jd f), b2), (1)))
	            ((s, (ch1, ch2)), ((b1', b2), (1))))
 | S2_jd : (s:state; ch1,ch2:codeheap; f:lab; b1,b2':iseq)
           (lookupC ch2 f b2')
           -> (Step ((s, (ch1, ch2)), ((b1, (jd f)), (2)))
	            ((s, (ch1, ch2)), ((b1, b2'), (2)))).

(****** Inference Rules ******)

Inductive Infer : tspec -> assert -> guar -> iseq -> Prop :=
 | I_seq : (t:tspec; inv,p,p':assert; g,g':guar; c:comm; b:iseq)
           ((tgetI t) == inv)
	    -> ((s,s1:state) (s1 = (Next c s))
	                  -> ((inv s) /\ (p s))
	                  -> ((inv s1)
	                   /\ (p' s1)
	                   /\ ((s':state) (g' s1 s') -> (g s s'))))
	    -> (Infer t p' g' b)
	    -> (Infer t p g (seq c b))
 | I_ld : (t:tspec; inv,p,p':assert; g,g':guar)
	  (rd,rs:reg; w:nat; b:iseq)
	  ((tgetI t) == inv)
	   -> ((H:heap; R:rfile) ((inv (H, R)) /\ (p (H, R)))
	                      -> (EX n:hval |
	      (((hread H (plus w (lookupR R rs))) = (hvsome n))
	       /\ (inv (H, (updateR R rd n)))
	       /\ (p' (H, (updateR R rd n)))
	       /\ ((s:state) (g' (H, (updateR R rd n)) s) -> (g (H, R) s)))))
	   -> (Infer t p' g' b)
	   -> (Infer t p g (ld rd rs w b))
 | I_st : (t:tspec; inv,p,p':assert; g,g':guar)
	  (rd,rs:reg; w:nat; b:iseq)
          ((tgetI t) == inv)
           -> ((H:heap; R:rfile) ((inv (H, R)) /\ (p (H, R)))
	       -> ((hindom (plus w (lookupR R rd)) H)
	       /\ (H1:heap)
	          (((hwrite H (plus w (lookupR R rd)) (lookupR R rs)) = H1)
	            -> (((inv (H1, R)) /\ (p' (H1, R)))
	            /\ ((s:state) (g' (H1, R) s) -> (g (H, R) s))))))
	   -> (Infer t p' g' b)
	   -> (Infer t p g (st rd w rs b))
 | I_bgt : (t:tspec; psi:codeheapty; inv,p,p',p'':assert; g,g',g'':guar)
	    (rs,rt:reg; f:lab; b:iseq)
            (((tgetCT t) == psi) /\ ((tgetI t) == inv))
	     -> (lookupCT psi f p' g')
	     -> ((H:heap; R:rfile)
	         (gt (lookupR R rs) (lookupR R rt))
		  -> ((inv (H, R)) /\ (p (H, R)))
		  -> ((p' (H, R))
		   /\ ((s:state) (g' (H, R) s) -> (g (H, R) s))))
	     -> ((H:heap; R:rfile)
	         (le (lookupR R rs) (lookupR R rt))
		  -> ((inv (H, R)) /\ (p (H, R)))
		  -> ((p'' (H, R))
		   /\ ((s:state) (g'' (H, R) s) -> (g (H, R) s))))
	     -> (Infer t p'' g'' b)
	     -> (Infer t p g (bgt rs rt f b))
 | I_be : (t:tspec; psi:codeheapty; inv,p,p',p'':assert; g,g',g'':guar)
	    (rs,rt:reg; f:lab; b:iseq)
            (((tgetCT t) == psi) /\ ((tgetI t) == inv))
	     -> (lookupCT psi f p' g')
	     -> ((H:heap; R:rfile)
	         ((lookupR R rs) = (lookupR R rt))
		  -> ((inv (H, R)) /\ (p (H, R)))
		  -> ((p' (H, R))
		   /\ ((s:state) (g' (H, R) s) -> (g (H, R) s))))
	     -> ((H:heap; R:rfile)
	         ~((lookupR R rs) = (lookupR R rt))
		  -> ((inv (H, R)) /\ (p (H, R)))
		  -> ((p'' (H, R))
		   /\ ((s:state) (g'' (H, R) s) -> (g (H, R) s))))
	     -> (Infer t p'' g'' b)
	     -> (Infer t p g (be rs rt f b))
 | I_yld : (t:tspec; inv,p:assert; g,A,G:guar; b:iseq)
            (((tgetI t) == inv) /\ ((tgetAG t) == (pairT A G)))
	     -> ((s:state) ((inv s) /\ (p s)) -> (g s s))
	     -> ((s:state) ((inv s) /\ (p s)) -> (s':state)(A s s') -> (p s'))
	     -> (Infer t p G b)
	     -> (Infer t p g (yield b))
 | I_jd : (t:tspec; psi:codeheapty; inv,p,p':assert; g,g':guar; f:lab)
            (((tgetCT t) == psi) /\ ((tgetI t) == inv))
	     -> (lookupCT psi f p' g')
	     -> ((s:state) ((inv s) /\ (p s))
	                -> ((p' s)
	                 /\ ((s':state) (g' s s') -> (g s s'))))
	     -> (Infer t p g (jd f)).

Inductive WF_CH_Aux : tspec -> codeheapty -> codeheap -> Prop :=
 | wf_c_emp : (t:tspec) (WF_CH_Aux t emptyCT emptyC)
 | wf_c_cons : (t:tspec; psi:codeheapty; ch:codeheap)
               (f:lab; p:assert; g:guar; b:iseq)
               (Infer t p g b)
               -> (WF_CH_Aux t psi ch)
               -> (WF_CH_Aux t (consCT f p g psi) (consC f b ch)).

Inductive WF_CH : tspec -> codeheap -> Prop :=
 | wf_c : (t:tspec; ch:codeheap)
           (WF_CH_Aux t (tgetCT t) ch)
	    -> (WF_CH t ch).

Inductive WFW : spec -> assert -> assert -> guar -> prog -> Prop :=
 | wf1_w : (phi:spec; t1,t2:tspec; inv,p1,p2:assert; g,A2,G2:guar)
           (s:state; ch1,ch2:codeheap; b1,b2:iseq)
           (((getInv phi) == inv) /\ ((getAG2 phi) == (pairT A2 G2)))
	    -> (((getT1 phi) == t1) /\ ((getT2 phi) == t2))
            -> (WF_CH t1 ch1)
	    -> (WF_CH t2 ch2)
	    -> ((inv s) /\ (p1 s))
	    -> ((s'':state) (g s s'') -> (p2 s''))
	    -> ((s':state) ((inv s') /\ (p2 s')) -> (s'':state) (A2 s' s'') -> (p2 s''))
	    -> (Infer t1 p1 g b1)
	    -> (Infer t2 p2 G2 b2)
            -> (WFW phi p1 p2 g ((s, (ch1, ch2)), ((b1,b2), (1))))
 | wf2_w : (phi:spec; t1,t2:tspec; inv,p1,p2:assert; g,A1,G1:guar)
           (s:state; ch1,ch2:codeheap; b1,b2:iseq)
           (((getInv phi) == inv) /\ ((getAG1 phi) == (pairT A1 G1)))
	    -> (((getT1 phi) == t1) /\ ((getT2 phi) == t2))
            -> (WF_CH t1 ch1)
	    -> (WF_CH t2 ch2)
	    -> ((inv s) /\ (p2 s))
	    -> ((s'':state) (g s s'') -> (p1 s''))
	    -> ((s':state) ((inv s') /\ (p1 s')) -> (s'':state) (A1 s' s'') -> (p1 s''))
	    -> (Infer t2 p2 g b2)
	    -> (Infer t1 p1 G1 b1)
            -> (WFW phi p1 p2 g ((s, (ch1, ch2)), ((b1,b2), (2)))).
